# templates/bar_templates.py

#BAR_JOB_STRUCT = """
#typedef struct {
#    uint8_t child_index;
#    int value;
#} {job_struct};
#"""

# REMOVED: BAR_JOB_CALLBACK outer shell — logic moves inline to setter
# KEPT: _exec_cb as a static LVGL animation helper, renamed to be standalone

BAR_CALLBACK = """
static void ${cb_name}_exec_cb(void *obj, int32_t v)
{
    lv_bar_set_value(obj, v, LV_ANIM_OFF);
}
"""

BAR_SETTER = """
void ${fn_name}(int value, uint32_t duration_ms)
{
    ui_child_t *c = &${screen_var}.children[${child_index}];
    if (!c->lv_obj || c->type != UI_CHILD_BAR)
        return;
    if (duration_ms == 0)
    {
        lv_bar_set_value(c->lv_obj, value, LV_ANIM_OFF);
        return;
    }
    lv_anim_t a;
    lv_anim_init(&a);
    lv_anim_set_var(&a, c->lv_obj);
    lv_anim_set_exec_cb(&a, ${cb_name}_exec_cb);
    lv_anim_set_values(&a, lv_bar_get_value(c->lv_obj), value);
    lv_anim_set_time(&a, duration_ms);
    lv_anim_start(&a);
}
"""

BAR_INIT = """
    case UI_CHILD_BAR:
        c->lv_obj = lv_bar_create(${screen_var}.lv_screen);
        lv_obj_set_pos(c->lv_obj, c->x, c->y);
        lv_obj_set_size(c->lv_obj, c->w, c->h);
        lv_bar_set_range(c->lv_obj, 0, 100);
        lv_bar_set_value(c->lv_obj, c->data.bar.value, LV_ANIM_OFF);
        break;
"""


# templates/icon_templates.py

#ICON_JOB_STRUCT = """
#typedef struct {
#    uint8_t child_index;
#    uint8_t state;
#{ {job_struct};
#"""

# REMOVED: IMAGE_JOB_CALLBACK — logic moves inline to setter
IMAGE_CALLBACK = ""

IMAGE_SETTER = """
void ${fn_name}(void)
{
    ui_child_t *c = &${screen_var}.children[${child_index}];
    if (c->type != UI_CHILD_IMAGE || c->lv_obj == NULL)
        return;
    c->data.image.src = &${child_id};
    lv_image_set_src(c->lv_obj, c->data.image.src);
}
"""

IMAGE_INIT = """
    case UI_CHILD_IMAGE:
        c->lv_obj = lv_image_create(${screen_var}.lv_screen);
        lv_obj_set_pos(c->lv_obj, c->x, c->y);
        if(c->data.image.src)
            lv_image_set_src(c->lv_obj, c->data.image.src);
        break;
"""


#include "ui_{screen_name}.h"
#include "ui_defs.h"
#include "assets.h"
#include "ui_worker.h"
#include "assets.h"
#include <stdio.h>

ui_screen_t ${screen_var} = {
    .name = "{screen_name}",
    .child_count = {child_count},
    .children = {
        {struct_entries}
    },
    .lv_screen = NULL
};

{callbacks}

{setters}

void ui_{screen_name}_init(void)
{
    {screen_var}.lv_screen = lv_obj_create(NULL);

    for (int i = 0; i < {child_count}; i++)
    {
        ui_child_t* c = &${screen_var}.children[i];

        switch (c->type)
        {
            {inits}
        }
    }

    ui_worker_init();
}


# core/context.py

class GenerationContext:
    def __init__(self):
        self.struct_entries = []
        self.callback_blocks = []
        self.setter_blocks = []
        self.init_blocks = []

    def add_struct(self, block):
        self.struct_entries.append(block)

    def add_callback(self, block):
        self.callback_blocks.append(block)

    def add_setter(self, block):
        self.setter_blocks.append(block)

    def add_init(self, block):
        self.init_blocks.append(block)

    def render_struct_entries(self):
        return "\n".join(self.struct_entries)

    def render_callbacks(self):
        return "\n".join(self.callback_blocks)

    def render_setters(self):
        return "\n".join(self.setter_blocks)

    def render_inits(self):
        return "\n".join(self.init_blocks)


#!/usr/bin/env python3
import os
import logging
import argparse
import subprocess
from os import path
from enum import Enum
from typing import List
from pathlib import Path

try:
    import png
except ImportError:
    raise ImportError("Need pypng package, do `pip3 install pypng`")

try:
    import lz4.block
except ImportError:
    raise ImportError("Need lz4 package, do `pip3 install lz4`")


def uint8_t(val) -> bytes:
    return val.to_bytes(1, byteorder='little')


def uint16_t(val) -> bytes:
    return val.to_bytes(2, byteorder='little')

def swap_uint16_t(val) -> bytes:
    return val.to_bytes(2, byteorder='big')

def uint24_t(val) -> bytes:
    return val.to_bytes(3, byteorder='little')


def uint32_t(val) -> bytes:
    try:
        return val.to_bytes(4, byteorder='little')
    except OverflowError:
        raise ParameterError(f"overflow: {hex(val)}")


def color_pre_multiply(r, g, b, a, background):
    bb = background & 0xff
    bg = (background >> 8) & 0xff
    br = (background >> 16) & 0xff

    return ((r * a + (255 - a) * br) >> 8, (g * a + (255 - a) * bg) >> 8,
            (b * a + (255 - a) * bb) >> 8, a)


class Error(Exception):

    def __str__(self):
        return self.__class__.__name__ + ': ' + ' '.join(self.args)


class FormatError(Error):
    """
    Problem with input filename format.
    BIN filename does not conform to standard lvgl bin image format
    """


class ParameterError(Error):
    """
    Parameter for LVGL image not correct
    """


class PngQuant:
    """
    Compress PNG file to 8bit mode using `pngquant`
    """

    def __init__(self, ncolors=256, dither=True, exec_path="") -> None:
        executable = path.join(exec_path, "pngquant")
        self.cmd = (f"{executable} {'--nofs' if not dither else ''} "
                    f"{ncolors}  --force - < ")

    def convert(self, filename) -> bytes:
        if not os.path.isfile(filename):
            raise BaseException(f"file not found: {filename}")

        try:
            compressed = subprocess.check_output(
                f'{self.cmd} "{str(filename)}"',
                stderr=subprocess.STDOUT,
                shell=True)
        except subprocess.CalledProcessError:
            raise BaseException(
                "cannot find pngquant tool, install it via "
                "`sudo apt install pngquant` for debian "
                "or `brew install pngquant` for macintosh "
                "For windows, you may need to download pngquant.exe from "
                "https://pngquant.org/, and put it in your PATH.")

        return compressed


class CompressMethod(Enum):
    NONE = 0x00
    RLE = 0x01
    LZ4 = 0x02


class ColorFormat(Enum):
    UNKNOWN = 0x00
    RAW = 0x01,
    RAW_ALPHA = 0x02,
    L8 = 0x06
    I1 = 0x07
    I2 = 0x08
    I4 = 0x09
    I8 = 0x0A
    A1 = 0x0B
    A2 = 0x0C
    A4 = 0x0D
    A8 = 0x0E
    AL88 = 0x15
    ARGB8888 = 0x10
    XRGB8888 = 0x11
    RGB565 = 0x12
    RGB565_SWAPPED = 0x1B
    ARGB8565 = 0x13
    RGB565A8 = 0x14
    RGB888 = 0x0F
    ARGB8888_PREMULTIPLIED = 0x1A

    @property
    def bpp(self) -> int:
        """
        Return bit per pixel for this cf
        """
        cf_map = {
            ColorFormat.L8: 8,
            ColorFormat.I1: 1,
            ColorFormat.I2: 2,
            ColorFormat.I4: 4,
            ColorFormat.I8: 8,
            ColorFormat.A1: 1,
            ColorFormat.A2: 2,
            ColorFormat.A4: 4,
            ColorFormat.A8: 8,
            ColorFormat.AL88: 16,
            ColorFormat.ARGB8888: 32,
            ColorFormat.XRGB8888: 32,
            ColorFormat.RGB565: 16,
            ColorFormat.RGB565_SWAPPED: 16,
            ColorFormat.RGB565A8: 16,  # 16bpp + a8 map
            ColorFormat.ARGB8565: 24,
            ColorFormat.RGB888: 24,
            ColorFormat.ARGB8888_PREMULTIPLIED: 32,
        }

        return cf_map[self] if self in cf_map else 0

    @property
    def ncolors(self) -> int:
        """
        Return number of colors in palette if cf is indexed1/2/4/8.
        Return zero if cf is not indexed format
        """

        cf_map = {
            ColorFormat.I1: 2,
            ColorFormat.I2: 4,
            ColorFormat.I4: 16,
            ColorFormat.I8: 256,
        }
        return cf_map.get(self, 0)

    @property
    def is_indexed(self) -> bool:
        """
        Return if cf is indexed color format
        """
        return self.ncolors != 0

    @property
    def is_alpha_only(self) -> bool:
        return ColorFormat.A1.value <= self.value <= ColorFormat.A8.value

    @property
    def has_alpha(self) -> bool:
        return self.is_alpha_only or self.is_indexed or self in (
            ColorFormat.AL88,
            ColorFormat.ARGB8888,
            ColorFormat.XRGB8888,  # const alpha: 0xff
            ColorFormat.ARGB8565,
            ColorFormat.RGB565A8,
            ColorFormat.ARGB8888_PREMULTIPLIED)

    @property
    def is_colormap(self) -> bool:
        return self in (ColorFormat.ARGB8888, ColorFormat.RGB888,
                        ColorFormat.XRGB8888, ColorFormat.RGB565A8,
                        ColorFormat.ARGB8565, ColorFormat.RGB565,
                        ColorFormat.RGB565_SWAPPED,
                        ColorFormat.ARGB8888_PREMULTIPLIED)

    @property
    def is_luma_only(self) -> bool:
        return self in (ColorFormat.L8, )


def bit_extend(value, bpp):
    """
    Extend value from bpp to 8 bit with interpolation to reduce rounding error.
    """

    if value == 0:
        return 0

    res = value
    bpp_now = bpp
    while bpp_now < 8:
        res |= value << (8 - bpp_now)
        bpp_now += bpp

    return res


def unpack_colors(data: bytes, cf: ColorFormat, w) -> List:
    """
    Unpack lvgl 1/2/4/8/16/32 bpp color to png color: alpha map, grey scale,
    or R,G,B,(A) map
    """
    ret = []
    bpp = cf.bpp
    if bpp == 8:
        ret = data
    elif bpp == 4:
        if cf == ColorFormat.A4:
            values = [x * 17 for x in range(16)]
        else:
            values = [x for x in range(16)]

        for p in data:
            for i in range(2):
                ret.append(values[(p >> (4 - i * 4)) & 0x0f])
                if len(ret) % w == 0:
                    break

    elif bpp == 2:
        if cf == ColorFormat.A2:
            values = [x * 85 for x in range(4)]
        else:  # must be ColorFormat.I2
            values = [x for x in range(4)]
        for p in data:
            for i in range(4):
                ret.append(values[(p >> (6 - i * 2)) & 0x03])
                if len(ret) % w == 0:
                    break
    elif bpp == 1:
        if cf == ColorFormat.A1:
            values = [0, 255]
        else:
            values = [0, 1]
        for p in data:
            for i in range(8):
                ret.append(values[(p >> (7 - i)) & 0x01])
                if len(ret) % w == 0:
                    break
    elif bpp == 16:
        if cf == ColorFormat.RGB565:
            #  This is RGB565
            pixels = [(data[2 * i + 1] << 8) | data[2 * i]
                    for i in range(len(data) // 2)]
            for p in pixels:
                ret.append(bit_extend((p >> 11) & 0x1f, 5))  # R
                ret.append(bit_extend((p >> 5) & 0x3f, 6))  # G
                ret.append(bit_extend((p >> 0) & 0x1f, 5))  # B
        elif cf == ColorFormat.RGB565_SWAPPED:
            #  This is RGB565_SWAPPED
            pixels = [(data[2 * i] << 8) | data[2 * i + 1]
                    for i in range(len(data) // 2)]
            for p in pixels:
                ret.append(bit_extend((p >> 11) & 0x1f, 5))  # R
                ret.append(bit_extend((p >> 5) & 0x3f, 6))  # G
                ret.append(bit_extend((p >> 0) & 0x1f, 5))  # B
        elif cf == ColorFormat.AL88:
            # AL88: low 8bit = Luminance, high 8bit = Alpha
            L = data[0::2]  # low byte: luminance
            A = data[1::2]  # high byte: alpha
            for luma, alpha in zip(L, A):
                ret.append(luma)  # L
                ret.append(alpha)  # A

    elif bpp == 24:
        if cf == ColorFormat.RGB888:
            B = data[0::3]
            G = data[1::3]
            R = data[2::3]
            for r, g, b in zip(R, G, B):
                ret += [r, g, b]
        elif cf == ColorFormat.RGB565A8:
            alpha_size = len(data) // 3
            pixel_alpha = data[-alpha_size:]
            pixel_data = data[:-alpha_size]
            pixels = [(pixel_data[2 * i + 1] << 8) | pixel_data[2 * i]
                      for i in range(len(pixel_data) // 2)]

            for a, p in zip(pixel_alpha, pixels):
                ret.append(bit_extend((p >> 11) & 0x1f, 5))  # R
                ret.append(bit_extend((p >> 5) & 0x3f, 6))  # G
                ret.append(bit_extend((p >> 0) & 0x1f, 5))  # B
                ret.append(a)
        elif cf == ColorFormat.ARGB8565:
            L = data[0::3]
            H = data[1::3]
            A = data[2::3]

            for h, l, a in zip(H, L, A):
                p = (h << 8) | (l)
                ret.append(bit_extend((p >> 11) & 0x1f, 5))  # R
                ret.append(bit_extend((p >> 5) & 0x3f, 6))  # G
                ret.append(bit_extend((p >> 0) & 0x1f, 5))  # B
                ret.append(a)  # A

    elif bpp == 32:
        B = data[0::4]
        G = data[1::4]
        R = data[2::4]
        A = data[3::4]
        if cf == ColorFormat.ARGB8888_PREMULTIPLIED:
            for r, g, b, a in zip(R, G, B, A):
                r = (r * a // 255)
                g = (g * a // 255)
                b = (b * a // 255)
                ret += [r, g, b, a]
        else:
            for r, g, b, a in zip(R, G, B, A):
                ret += [r, g, b, a]
    else:
        assert 0

    return ret


def write_c_array_file(
        w: int, h: int,
        stride: int,
        cf: ColorFormat,
        filename: str,
        outputname: str,
        premultiplied: bool,
        compress: CompressMethod,
        data: bytes):
    varname = path.basename(filename).split('.')[0].replace("-", "_").replace(".", "_") if outputname is None else outputname

    flags = "0"
    if compress is not CompressMethod.NONE:
        flags += " | LV_IMAGE_FLAGS_COMPRESSED"
    if premultiplied:
        flags += " | LV_IMAGE_FLAGS_PREMULTIPLIED"

    macro = "LV_ATTRIBUTE_" + varname.upper()
    header = f'''
#if defined(LV_LVGL_H_INCLUDE_SIMPLE)
#include "lvgl.h"
#elif defined(LV_LVGL_H_INCLUDE_SYSTEM)
#include <lvgl.h>
#elif defined(LV_BUILD_TEST)
#include "../lvgl.h"
#else
#include "lvgl.h"
#endif

#ifndef LV_ATTRIBUTE_MEM_ALIGN
#define LV_ATTRIBUTE_MEM_ALIGN
#endif

#ifndef {macro}
#define {macro}
#endif

static const
LV_ATTRIBUTE_MEM_ALIGN LV_ATTRIBUTE_LARGE_CONST {macro}
uint8_t {varname}_map[] = {{
'''

    ending = f'''
}};

const lv_image_dsc_t {varname} = {{
  .header = {{
    .magic = LV_IMAGE_HEADER_MAGIC,
    .cf = LV_COLOR_FORMAT_{cf.name},
    .flags = {flags},
    .w = {w},
    .h = {h},
    .stride = {stride},
    .reserved_2 = 0,
  }},
  .data_size = sizeof({varname}_map),
  .data = {varname}_map,
  .reserved = NULL,
}};

'''

    def write_binary(f, data, stride):
        stride = 16 if stride == 0 else stride
        for i, v in enumerate(data):
            if i % stride == 0:
                f.write("\n    ")
            f.write(f"0x{v:02x},")
        f.write("\n")

    with open(filename, "w+") as f:
        f.write(header)

        if compress != CompressMethod.NONE:
            write_binary(f, data, 16)
        else:
            # write palette separately
            ncolors = cf.ncolors
            if ncolors:
                write_binary(f, data[:ncolors * 4], 16)

            write_binary(f, data[ncolors * 4:], stride)

        f.write(ending)


class LVGLImageHeader:

    def __init__(self,
                 cf: ColorFormat = ColorFormat.UNKNOWN,
                 w: int = 0,
                 h: int = 0,
                 stride: int = 0,
                 align: int = 1,
                 flags: int = 0):
        self.cf = cf
        self.flags = flags
        self.w = w & 0xffff
        self.h = h & 0xffff
        if w > 0xffff or h > 0xffff:
            raise ParameterError(f"w, h overflow: {w}x{h}")
        if align < 1:
            # stride align in bytes must be larger than 1
            raise ParameterError(f"Invalid stride align: {align}")

        self.stride = self.stride_align(align) if stride == 0 else stride

    def stride_align(self, align: int) -> int:
        stride = self.stride_default
        if align == 1:
            pass
        elif align > 1:
            stride = (stride + align - 1) // align
            stride *= align
        else:
            raise ParameterError(f"Invalid stride align: {align}")

        self.stride = stride
        return stride

    @property
    def stride_default(self) -> int:
        return (self.w * self.cf.bpp + 7) // 8

    @property
    def binary(self) -> bytearray:
        binary = bytearray()
        binary += uint8_t(0x19)  # magic number for lvgl version 9
        binary += uint8_t(self.cf.value)
        binary += uint16_t(self.flags)  # 16bits flags

        binary += uint16_t(self.w)  # 16bits width
        binary += uint16_t(self.h)  # 16bits height
        binary += uint16_t(self.stride)  # 16bits stride

        binary += uint16_t(0)  # 16bits reserved
        return binary

    def from_binary(self, data: bytes):
        if len(data) < 12:
            raise FormatError("invalid header length")

        try:
            self.cf = ColorFormat(data[1] & 0x1f)  # color format
        except ValueError as exc:
            raise FormatError(f"invalid color format: {hex(data[0])}") from exc
        self.w = int.from_bytes(data[4:6], 'little')
        self.h = int.from_bytes(data[6:8], 'little')
        self.stride = int.from_bytes(data[8:10], 'little')
        return self


class LVGLCompressData:

    def __init__(self,
                 cf: ColorFormat,
                 method: CompressMethod,
                 raw_data: bytes = b''):
        self.blk_size = (cf.bpp + 7) // 8
        self.compress = method
        self.raw_data = raw_data
        self.raw_data_len = len(raw_data)
        self.compressed = self._compress(raw_data)

    def _compress(self, raw_data: bytes) -> bytearray:
        if self.compress == CompressMethod.NONE:
            return raw_data

        if self.compress == CompressMethod.RLE:
            # RLE compression performs on pixel unit, pad data to pixel unit
            pad = b'\x00' * 0
            if self.raw_data_len % self.blk_size:
                pad = b'\x00' * (self.blk_size - self.raw_data_len % self.blk_size)
            compressed = RLEImage().rle_compress(raw_data + pad, self.blk_size)
        elif self.compress == CompressMethod.LZ4:
            compressed = lz4.block.compress(raw_data, store_size=False)
        else:
            raise ParameterError(f"Invalid compress method: {self.compress}")

        self.compressed_len = len(compressed)

        bin = bytearray()
        bin += uint32_t(self.compress.value)
        bin += uint32_t(self.compressed_len)
        bin += uint32_t(self.raw_data_len)
        bin += compressed
        return bin


class LVGLImage:

    def __init__(self,
                 cf: ColorFormat = ColorFormat.UNKNOWN,
                 w: int = 0,
                 h: int = 0,
                 data: bytes = b'') -> None:
        self.stride = 0  # default no valid stride value
        self.premultiplied = False
        self.rgb565_dither = False
        self.nema_gfx = False
        self.set_data(cf, w, h, data)

    def __repr__(self) -> str:
        return (f"'LVGL image {self.w}x{self.h}, {self.cf.name}, "
                f"{'Pre-multiplied, ' if self.premultiplied else ''}"
                f"stride: {self.stride} "
                f"(12+{self.data_len})Byte'")

    def adjust_stride(self, stride: int = 0, align: int = 1):
        """
        Stride can be set directly, or by stride alignment in bytes
        """
        if self.stride == 0:
            #  stride can only be 0, when LVGLImage is created with empty data
            logging.warning("Cannot adjust stride for empty image")
            return

        if align >= 1 and stride == 0:
            # The header with specified stride alignment
            header = LVGLImageHeader(self.cf, self.w, self.h, align=align)
            stride = header.stride
        elif stride > 0:
            pass
        else:
            raise ParameterError(f"Invalid parameter, align:{align},"
                                 f" stride:{stride}")

        if self.stride == stride:
            return  # no stride adjustment

        # if current image is empty, no need to do anything
        if self.data_len == 0:
            self.stride = 0
            return

        current = LVGLImageHeader(self.cf, self.w, self.h, stride=self.stride)

        if stride < current.stride_default:
            raise ParameterError(f"Stride is too small:{stride}, "
                                 f"minimal:{current.stride_default}")

        def change_stride(data: bytearray, h, current_stride, new_stride):
            data_in = data
            data_out = []  # stride adjusted new data
            if new_stride < current_stride:  # remove padding byte
                for i in range(h):
                    start = i * current_stride
                    end = start + new_stride
                    data_out.append(data_in[start:end])
            else:  # adding more padding bytes
                padding = b'\x00' * (new_stride - current_stride)
                for i in range(h):
                    data_out.append(data_in[i * current_stride:(i + 1) *
                                            current_stride])
                    data_out.append(padding)
            return b''.join(data_out)

        palette_size = self.cf.ncolors * 4
        data_out = [self.data[:palette_size]]
        data_out.append(
            change_stride(self.data[palette_size:], self.h, current.stride,
                          stride))

        # deal with alpha map for RGB565A8
        if self.cf == ColorFormat.RGB565A8:
            logging.warning("handle RGB565A8 alpha map")
            a8_stride = self.stride // 2
            a8_map = self.data[-a8_stride * self.h:]
            data_out.append(
                change_stride(a8_map, self.h, current.stride // 2,
                              stride // 2))

        self.stride = stride
        self.data = bytearray(b''.join(data_out))

    def premultiply(self):
        """
        Pre-multiply image RGB data with alpha, set corresponding image header flags
        """
        if self.premultiplied:
            raise ParameterError("Image already pre-multiplied")

        if not self.cf.has_alpha:
            raise ParameterError(f"Image has no alpha channel: {self.cf.name}")

        if self.cf.is_indexed:

            def multiply(b, g, r, a):
                # The precision is reduced, the correct way would be to divide by 255,
                # but this is consistent with the premultiply function in the code.
                r, g, b = (r * a) >> 8, (g * a) >> 8, (b * a) >> 8
                return uint8_t(b) + uint8_t(g) + uint8_t(r) + uint8_t(a)

            # process the palette only.
            palette_size = self.cf.ncolors * 4
            palette = self.data[:palette_size]
            palette = [
                multiply(palette[i], palette[i + 1], palette[i + 2],
                         palette[i + 3]) for i in range(0, len(palette), 4)
            ]
            palette = b''.join(palette)
            self.data = palette + self.data[palette_size:]
        elif self.cf is ColorFormat.ARGB8888:

            def multiply(b, g, r, a):
                r, g, b = (r * a) >> 8, (g * a) >> 8, (b * a) >> 8
                return uint32_t((a << 24) | (r << 16) | (g << 8) | (b << 0))

            line_width = self.w * 4
            for h in range(self.h):
                offset = h * self.stride
                map = self.data[offset:offset + self.stride]

                processed = b''.join([
                    multiply(map[i], map[i + 1], map[i + 2], map[i + 3])
                    for i in range(0, line_width, 4)
                ])
                self.data[offset:offset + line_width] = processed
        elif self.cf is ColorFormat.RGB565A8:

            def multiply(data, a):
                r = (data >> 11) & 0x1f
                g = (data >> 5) & 0x3f
                b = (data >> 0) & 0x1f

                r, g, b = (r * a) // 255, (g * a) // 255, (b * a) // 255
                return uint16_t((r << 11) | (g << 5) | (b << 0))

            line_width = self.w * 2
            for h in range(self.h):
                # alpha map offset for this line
                offset = self.h * self.stride + h * (self.stride // 2)
                a = self.data[offset:offset + self.stride // 2]

                # RGB map offset
                offset = h * self.stride
                rgb = self.data[offset:offset + self.stride]

                processed = b''.join([
                    multiply((rgb[i + 1] << 8) | rgb[i], a[i // 2])
                    for i in range(0, line_width, 2)
                ])
                self.data[offset:offset + line_width] = processed
        elif self.cf is ColorFormat.ARGB8565:

            def multiply(data, a):
                r = (data >> 11) & 0x1f
                g = (data >> 5) & 0x3f
                b = (data >> 0) & 0x1f

                r, g, b = (r * a) // 255, (g * a) // 255, (b * a) // 255
                return uint24_t((a << 16) | (r << 11) | (g << 5) | (b << 0))

            line_width = self.w * 3
            for h in range(self.h):
                offset = h * self.stride
                map = self.data[offset:offset + self.stride]

                processed = b''.join([
                    multiply((map[i + 1] << 8) | map[i], map[i + 2])
                    for i in range(0, line_width, 3)
                ])
                self.data[offset:offset + line_width] = processed
        else:
            raise ParameterError(f"Not supported yet: {self.cf.name}")

        self.premultiplied = True

    @property
    def data_len(self) -> int:
        """
        Return data_len in byte of this image, excluding image header
        """

        # palette is always in ARGB format, 4Byte per color
        p = self.cf.ncolors * 4 if self.is_indexed and self.w * self.h else 0
        p += self.stride * self.h
        if self.cf is ColorFormat.RGB565A8:
            a8_stride = self.stride // 2
            p += a8_stride * self.h
        return p

    @property
    def header(self) -> bytearray:
        return LVGLImageHeader(self.cf, self.w, self.h)

    @property
    def is_indexed(self):
        return self.cf.is_indexed

    def set_data(self,
                 cf: ColorFormat,
                 w: int,
                 h: int,
                 data: bytes,
                 stride: int = 0):
        """
        Directly set LVGL image parameters
        """

        if w > 0xffff or h > 0xffff:
            raise ParameterError(f"w, h overflow: {w}x{h}")

        self.cf = cf
        self.w = w
        self.h = h

        # if stride is 0, then it's aligned to 1byte by default,
        # let image header handle it
        self.stride = LVGLImageHeader(cf, w, h, stride, align=1).stride

        if self.data_len != len(data):
            raise ParameterError(f"{self} data length error got: {len(data)}, "
                                 f"expect: {self.data_len}, {self}")

        self.data = data

        return self

    def from_data(self, data: bytes):
        header = LVGLImageHeader().from_binary(data)
        return self.set_data(header.cf, header.w, header.h,
                             data[len(header.binary):], header.stride)

    def from_bin(self, filename: str):
        """
        Read from existing bin file and update image parameters
        """

        if not filename.endswith(".bin"):
            raise FormatError("filename not ended with '.bin'")

        with open(filename, "rb") as f:
            data = f.read()
            return self.from_data(data)

    def _check_ext(self, filename: str, ext):
        if not filename.lower().endswith(ext):
            raise FormatError(f"filename not ended with {ext}")

    def _check_dir(self, filename: str):
        dir = path.dirname(filename)
        if dir and not path.exists(dir):
            logging.info(f"mkdir of {dir} for {filename}")
            os.makedirs(dir)

    def to_bin(self,
               filename: str,
               compress: CompressMethod = CompressMethod.NONE):
        """
        Write this image to file, filename should be ended with '.bin'
        """
        self._check_ext(filename, ".bin")
        self._check_dir(filename)

        with open(filename, "wb+") as f:
            bin = bytearray()
            flags = 0
            flags |= 0x08 if compress != CompressMethod.NONE else 0
            flags |= 0x01 if self.premultiplied else 0

            header = LVGLImageHeader(self.cf,
                                     self.w,
                                     self.h,
                                     self.stride,
                                     flags=flags)
            bin += header.binary
            compressed = LVGLCompressData(self.cf, compress, self.data)
            bin += compressed.compressed

            f.write(bin)

        return self

    def to_c_array(self,
                   filename: str,
                   compress: CompressMethod = CompressMethod.NONE,
                   outputname: str = None):
        self._check_ext(filename, ".c")
        self._check_dir(filename)

        if compress != CompressMethod.NONE:
            data = LVGLCompressData(self.cf, compress, self.data).compressed
        else:
            data = self.data
        write_c_array_file(self.w, self.h, self.stride, self.cf, filename, outputname,
                           self.premultiplied,
                           compress, data)

    def to_png(self, filename: str):
        self._check_ext(filename, ".png")
        self._check_dir(filename)

        old_stride = self.stride
        self.adjust_stride(align=1)
        if self.cf.is_indexed:
            data = self.data
            # Separate lvgl bin image data to palette and bitmap
            # The palette is in format of [(RGBA), (RGBA)...].
            # LVGL palette is in format of B,G,R,A,...
            palette = [(data[i * 4 + 2], data[i * 4 + 1], data[i * 4 + 0],
                        data[i * 4 + 3]) for i in range(self.cf.ncolors)]

            data = data[self.cf.ncolors * 4:]

            encoder = png.Writer(self.w,
                                 self.h,
                                 palette=palette,
                                 bitdepth=self.cf.bpp)
            # separate packed data to plain data
            data = unpack_colors(data, self.cf, self.w)
        elif self.cf.is_alpha_only:
            # separate packed data to plain data
            transparency = unpack_colors(self.data, self.cf, self.w)
            data = []
            for a in transparency:
                data += [0, 0, 0, a]
            encoder = png.Writer(self.w, self.h, greyscale=False, alpha=True)
        elif self.cf == ColorFormat.L8:
            # to grayscale
            encoder = png.Writer(self.w,
                                 self.h,
                                 bitdepth=self.cf.bpp,
                                 greyscale=True,
                                 alpha=False)
            data = self.data
        elif self.cf == ColorFormat.AL88:
            # to grayscale with alpha
            encoder = png.Writer(self.w,
                                 self.h,
                                 bitdepth=8,
                                 greyscale=True,
                                 alpha=True)
            data = unpack_colors(self.data, self.cf, self.w)
        elif self.cf.is_colormap:
            encoder = png.Writer(self.w,
                                 self.h,
                                 alpha=self.cf.has_alpha,
                                 greyscale=False)
            data = unpack_colors(self.data, self.cf, self.w)
        else:
            logging.warning(f"missing logic: {self.cf.name}")
            return

        with open(filename, "wb") as f:
            encoder.write_array(f, data)

        self.adjust_stride(stride=old_stride)

    def from_png(self,
                 filename: str,
                 cf: ColorFormat = None,
                 background: int = 0x00_00_00,
                 rgb565_dither=False,
                 nema_gfx=False):
        """
        Create lvgl image from png file.
        If cf is none, used I1/2/4/8 based on palette size
        """

        self.background = background
        self.rgb565_dither = rgb565_dither
        self.nema_gfx = nema_gfx

        if cf is None:  # guess cf from filename
            # split filename string and match with ColorFormat to check
            # which cf to use
            names = str(path.basename(filename)).split(".")
            for c in names[1:-1]:
                if c in ColorFormat.__members__:
                    cf = ColorFormat[c]
                    break

        if cf is None or cf.is_indexed:  # palette mode
            self._png_to_indexed(cf, filename)
        elif cf.is_alpha_only:
            self._png_to_alpha_only(cf, filename)
        elif cf == ColorFormat.AL88:
            self._png_to_al88(cf, filename)
        elif cf.is_luma_only:
            self._png_to_luma_only(cf, filename)
        elif cf.is_colormap:
            self._png_to_colormap(cf, filename)
        else:
            logging.warning(f"missing logic: {cf.name}")

        logging.info(f"from png: {filename}, cf: {self.cf.name}")
        return self

    def _png_to_indexed(self, cf: ColorFormat, filename: str):
        # convert to palette mode
        auto_cf = cf is None

        # read the image data to get the metadata
        reader = png.Reader(filename=filename)
        w, h, rows, metadata = reader.read()

        # to preserve original palette data only convert the image if needed. For this
        # check if image has a palette and the requested palette size equals the existing one
        if not 'palette' in metadata or not auto_cf and len(metadata['palette']) !=  2 ** cf.bpp:
            # reread and convert file
            reader = png.Reader(
                bytes=PngQuant(256 if auto_cf else cf.ncolors).convert(filename))
            w, h, rows, _ = reader.read()

        palette = reader.palette(alpha="force")  # always return alpha

        palette_len = len(palette)
        if auto_cf:
            if palette_len <= 2:
                cf = ColorFormat.I1
            elif palette_len <= 4:
                cf = ColorFormat.I2
            elif palette_len <= 16:
                cf = ColorFormat.I4
            else:
                cf = ColorFormat.I8

        if palette_len != cf.ncolors:
            if not auto_cf:
                logging.warning(
                    f"{path.basename(filename)} palette: {palette_len}, "
                    f"extended to: {cf.ncolors}")
            palette += [(255, 255, 255, 0)] * (cf.ncolors - palette_len)

        # Assemble lvgl image palette from PNG palette.
        # PNG palette is a list of tuple(R,G,B,A)

        rawdata = bytearray()
        for (r, g, b, a) in palette:
            rawdata += uint32_t((a << 24) | (r << 16) | (g << 8) | (b << 0))

        # pack data if not in I8 format
        if cf == ColorFormat.I8:
            for e in rows:
                if self.nema_gfx:
                    e = bytearray((x >> 4) | ((x & 0x0F) << 4) for x in e)
                rawdata += e
        else:
            for e in png.pack_rows(rows, cf.bpp):
                rawdata += e

        self.set_data(cf, w, h, rawdata)

    def _png_to_alpha_only(self, cf: ColorFormat, filename: str):
        reader = png.Reader(str(filename))
        w, h, rows, info = reader.asRGBA8()
        if not info['alpha']:
            raise FormatError(f"{filename} has no alpha channel")

        rawdata = bytearray()
        if cf == ColorFormat.A8:
            for row in rows:
                A = row[3::4]
                for e in A:
                    rawdata += uint8_t(e)
        else:
            shift = 8 - cf.bpp
            mask = 2**cf.bpp - 1
            rows = [[(a >> shift) & mask for a in row[3::4]] for row in rows]
            for row in png.pack_rows(rows, cf.bpp):
                rawdata += row

        self.set_data(cf, w, h, rawdata)

    def sRGB_to_linear(self, x):
        if x < 0.04045:
            return x / 12.92
        return pow((x + 0.055) / 1.055, 2.4)

    def linear_to_sRGB(self, y):
        if y <= 0.0031308:
            return 12.92 * y
        return 1.055 * pow(y, 1 / 2.4) - 0.055

    def _png_to_al88(self, cf: ColorFormat, filename: str):
        reader = png.Reader(str(filename))
        w, h, rows, info = reader.asRGBA8()
        if not info['alpha']:
            raise FormatError(f"{filename} has no alpha channel")

        rawdata = bytearray()
        for row in rows:
            R = row[0::4]
            G = row[1::4]
            B = row[2::4]
            A = row[3::4]
            for r, g, b, a in zip(R, G, B, A):
                # Calculate luminance using ITU-R BT.709 coefficients
                r_linear = self.sRGB_to_linear(r / 255.0)
                g_linear = self.sRGB_to_linear(g / 255.0)
                b_linear = self.sRGB_to_linear(b / 255.0)
                luma = 0.2126 * r_linear + 0.7152 * g_linear + 0.0722 * b_linear
                luma_byte = int(self.linear_to_sRGB(luma) * 255)
                # AL88: low byte = luminance, high byte = alpha
                rawdata += uint8_t(luma_byte)  # L
                rawdata += uint8_t(a)          # A

        self.set_data(ColorFormat.AL88, w, h, rawdata)

    def _png_to_luma_only(self, cf: ColorFormat, filename: str):
        reader = png.Reader(str(filename))
        w, h, rows, info = reader.asRGBA8()
        rawdata = bytearray()
        for row in rows:
            R = row[0::4]
            G = row[1::4]
            B = row[2::4]
            A = row[3::4]
            for r, g, b, a in zip(R, G, B, A):
                r, g, b, a = color_pre_multiply(r, g, b, a, self.background)
                r = self.sRGB_to_linear(r / 255.0)
                g = self.sRGB_to_linear(g / 255.0)
                b = self.sRGB_to_linear(b / 255.0)
                luma = 0.2126 * r + 0.7152 * g + 0.0722 * b
                rawdata += uint8_t(int(self.linear_to_sRGB(luma) * 255))

        self.set_data(ColorFormat.L8, w, h, rawdata)

    def _png_to_colormap(self, cf, filename: str):

        if cf == ColorFormat.ARGB8888:

            def pack(r, g, b, a):
                return uint32_t((a << 24) | (r << 16) | (g << 8) | (b << 0))
        elif cf == ColorFormat.ARGB8888_PREMULTIPLIED:

            def pack(r, g, b, a):
                # Premultiply RGB by Alpha
                r = (r * a // 255)
                g = (g * a // 255)
                b = (b * a // 255)

                # Pack into ARGB8888 format
                return uint32_t((a << 24) | (r << 16) | (g << 8) | (b << 0))
        elif cf == ColorFormat.XRGB8888:

            def pack(r, g, b, a):
                r, g, b, a = color_pre_multiply(r, g, b, a, self.background)
                return uint32_t((0xff << 24) | (r << 16) | (g << 8) | (b << 0))
        elif cf == ColorFormat.RGB888:

            def pack(r, g, b, a):
                r, g, b, a = color_pre_multiply(r, g, b, a, self.background)
                return uint24_t((r << 16) | (g << 8) | (b << 0))
        elif cf == ColorFormat.RGB565:

            def pack(r, g, b, a):
                r, g, b, a = color_pre_multiply(r, g, b, a, self.background)
                color = (r >> 3) << 11
                color |= (g >> 2) << 5
                color |= (b >> 3) << 0
                return uint16_t(color)
        elif cf == ColorFormat.RGB565_SWAPPED:

            def pack(r, g, b, a):
                r, g, b, a = color_pre_multiply(r, g, b, a, self.background)
                color = (r >> 3) << 11
                color |= (g >> 2) << 5
                color |= (b >> 3) << 0
                return swap_uint16_t(color)

        elif cf == ColorFormat.RGB565A8:

            def pack(r, g, b, a):
                color = (r >> 3) << 11
                color |= (g >> 2) << 5
                color |= (b >> 3) << 0
                return uint16_t(color)
        elif cf == ColorFormat.ARGB8565:

            def pack(r, g, b, a):
                color = (r >> 3) << 11
                color |= (g >> 2) << 5
                color |= (b >> 3) << 0
                return uint24_t((a << 16) | color)
        else:
            raise FormatError(f"Invalid color format: {cf.name}")

        reader = png.Reader(str(filename))
        w, h, rows, _ = reader.asRGBA8()
        rawdata = bytearray()
        alpha = bytearray()
        for y, row in enumerate(rows):
            R = row[0::4]
            G = row[1::4]
            B = row[2::4]
            A = row[3::4]
            for x, (r, g, b, a) in enumerate(zip(R, G, B, A)):
                if cf == ColorFormat.RGB565A8:
                    alpha += uint8_t(a)

                if (
                    self.rgb565_dither and
                    cf in (ColorFormat.RGB565, ColorFormat.RGB565_SWAPPED, ColorFormat.RGB565A8, ColorFormat.ARGB8565)
                ):
                    threshold_id = ((y & 7) << 3) + (x & 7)

                    r = min(r + red_thresh[threshold_id], 0xFF) & 0xF8
                    g = min(g + green_thresh[threshold_id], 0xFF) & 0xFC
                    b = min(b + blue_thresh[threshold_id], 0xFF) & 0xF8

                rawdata += pack(r, g, b, a)

        if cf == ColorFormat.RGB565A8:
            rawdata += alpha

        self.set_data(cf, w, h, rawdata)


red_thresh = [
  1, 7, 3, 5, 0, 8, 2, 6,
  7, 1, 5, 3, 8, 0, 6, 2,
  3, 5, 0, 8, 2, 6, 1, 7,
  5, 3, 8, 0, 6, 2, 7, 1,
  0, 8, 2, 6, 1, 7, 3, 5,
  8, 0, 6, 2, 7, 1, 5, 3,
  2, 6, 1, 7, 3, 5, 0, 8,
  6, 2, 7, 1, 5, 3, 8, 0
]

green_thresh = [
  1, 3, 2, 2, 3, 1, 2, 2,
  2, 2, 0, 4, 2, 2, 4, 0,
  3, 1, 2, 2, 1, 3, 2, 2,
  2, 2, 4, 0, 2, 2, 0, 4,
  1, 3, 2, 2, 3, 1, 2, 2,
  2, 2, 0, 4, 2, 2, 4, 0,
  3, 1, 2, 2, 1, 3, 2, 2,
  2, 2, 4, 0, 2, 2, 0, 4
]

blue_thresh = [
  5, 3, 8, 0, 6, 2, 7, 1,
  3, 5, 0, 8, 2, 6, 1, 7,
  8, 0, 6, 2, 7, 1, 5, 3,
  0, 8, 2, 6, 1, 7, 3, 5,
  6, 2, 7, 1, 5, 3, 8, 0,
  2, 6, 1, 7, 3, 5, 0, 8,
  7, 1, 5, 3, 8, 0, 6, 2,
  1, 7, 3, 5, 0, 8, 2, 6
]


class RLEHeader:

    def __init__(self, blksize: int, len: int):
        self.blksize = blksize
        self.len = len

    @property
    def binary(self):
        magic = 0x5aa521e0

        rle_header = self.blksize
        rle_header |= (self.len & 0xffffff) << 4

        binary = bytearray()
        binary.extend(uint32_t(magic))
        binary.extend(uint32_t(rle_header))
        return binary


class RLEImage(LVGLImage):

    def __init__(self,
                 cf: ColorFormat = ColorFormat.UNKNOWN,
                 w: int = 0,
                 h: int = 0,
                 data: bytes = b'') -> None:
        super().__init__(cf, w, h, data)

    def to_rle(self, filename: str):
        """
        Compress this image to file, filename should be ended with '.rle'
        """
        self._check_ext(filename, ".rle")
        self._check_dir(filename)

        # compress image data excluding lvgl image header
        blksize = (self.cf.bpp + 7) // 8
        compressed = self.rle_compress(self.data, blksize)
        with open(filename, "wb+") as f:
            header = RLEHeader(blksize, len(self.data)).binary
            header.extend(self.header.binary)
            f.write(header)
            f.write(compressed)

    def rle_compress(self, data: bytearray, blksize: int, threshold=16):
        index = 0
        data_len = len(data)
        compressed_data = []
        memview = memoryview(data)
        while index < data_len:
            repeat_cnt = self.get_repeat_count(memview[index:], blksize)
            if repeat_cnt == 0:
                # done
                break
            elif repeat_cnt < threshold:
                nonrepeat_cnt = self.get_nonrepeat_count(
                    memview[index:], blksize, threshold)
                ctrl_byte = uint8_t(nonrepeat_cnt | 0x80)
                compressed_data.append(ctrl_byte)
                compressed_data.append(memview[index:index +
                                               nonrepeat_cnt * blksize])
                index += nonrepeat_cnt * blksize
            else:
                ctrl_byte = uint8_t(repeat_cnt)
                compressed_data.append(ctrl_byte)
                compressed_data.append(memview[index:index + blksize])
                index += repeat_cnt * blksize

        return b"".join(compressed_data)

    def get_repeat_count(self, data: bytearray, blksize: int):
        if len(data) < blksize:
            return 0

        start = data[:blksize]
        index = 0
        repeat_cnt = 0
        value = 0

        while index < len(data):
            value = data[index:index + blksize]

            if value == start:
                repeat_cnt += 1
                if repeat_cnt == 127:  # limit max repeat count to max value of signed char.
                    break
            else:
                break
            index += blksize

        return repeat_cnt

    def get_nonrepeat_count(self, data: bytearray, blksize: int, threshold):
        if len(data) < blksize:
            return 0

        pre_value = data[:blksize]

        index = 0
        nonrepeat_count = 0

        repeat_cnt = 0
        while True:
            value = data[index:index + blksize]
            if value == pre_value:
                repeat_cnt += 1
                if repeat_cnt > threshold:
                    # repeat found.
                    break
            else:
                pre_value = value
                nonrepeat_count += 1 + repeat_cnt
                repeat_cnt = 0
                if nonrepeat_count >= 127:  # limit max repeat count to max value of signed char.
                    nonrepeat_count = 127
                    break

            index += blksize  # move to next position
            if index >= len(data):  # data end
                nonrepeat_count += repeat_cnt
                break

        return nonrepeat_count


class RAWImage():
    '''
    RAW image is an exception to LVGL image, it has color format of RAW or RAW_ALPHA.
    It has same image header as LVGL image, but the data is pure raw data from file.
    It does not support stride adjustment etc. features for LVGL image.
    It only supports convert an image to C array with RAW or RAW_ALPHA format.
    '''
    CF_SUPPORTED = (ColorFormat.RAW, ColorFormat.RAW_ALPHA)

    class NotSupported(NotImplementedError):
        pass

    def __init__(self,
                 cf: ColorFormat = ColorFormat.UNKNOWN,
                 data: bytes = b'') -> None:
        self.cf = cf
        self.data = data

    def to_c_array(self,
                   filename: str,
                   outputname: str = None):
        # Image size is set to zero, to let PNG or JPEG decoder to handle it
        # Stride is meaningless for RAW image
        write_c_array_file(0, 0, 0, self.cf, filename, outputname,
                           False, CompressMethod.NONE, self.data)

    def from_file(self,
                  filename: str,
                  cf: ColorFormat = None):
        if cf not in RAWImage.CF_SUPPORTED:
            raise RAWImage.NotSupported(f"Invalid color format: {cf.name}")

        with open(filename, "rb") as f:
            self.data = f.read()
        self.cf = cf
        return self


class OutputFormat(Enum):
    C_ARRAY = "C"
    BIN_FILE = "BIN"
    PNG_FILE = "PNG"  # convert to lvgl image and then to png


class PNGConverter:

    def __init__(self,
                 files: List,
                 cf: ColorFormat,
                 ofmt: OutputFormat,
                 odir: str,
                 background: int = 0x00,
                 align: int = 1,
                 premultiply: bool = False,
                 compress: CompressMethod = CompressMethod.NONE,
                 keep_folder=True,
                 rgb565_dither=False,
                 nema_gfx=False) -> None:
        self.files = files
        self.cf = cf
        self.ofmt = ofmt
        self.output = odir
        self.pngquant = None
        self.keep_folder = keep_folder
        self.align = align
        self.premultiply = premultiply
        self.compress = compress
        self.background = background
        self.rgb565_dither = rgb565_dither
        self.nema_gfx = nema_gfx

    def _replace_ext(self, input, ext, outputname: str = None):
        if self.keep_folder:
            name, _ = path.splitext(input)
        else:
            name, _ = path.splitext(path.basename(input))

        # change output name to 'outputname', if specified
        if outputname is not None:
            name = path.join(path.dirname(name), outputname)

        output = name + ext
        output = path.join(self.output, output)
        return output

    def convert(self, outputname: str):
        if len(self.files) > 1 and outputname is not None:
            raise BaseException(f"Cannot specify output name when converting more than one file.")

        output = []
        for f in self.files:
            if self.cf in (ColorFormat.RAW, ColorFormat.RAW_ALPHA):
                # Process RAW image explicitly
                img = RAWImage().from_file(f, self.cf)
                img.to_c_array(self._replace_ext(f, ".c", outputname), outputname=outputname)
            else:
                img = LVGLImage().from_png(f, self.cf, background=self.background, rgb565_dither=self.rgb565_dither, nema_gfx=self.nema_gfx)
                img.adjust_stride(align=self.align)

                if self.premultiply:
                    img.premultiply()
                output.append((f, img))
                if self.ofmt == OutputFormat.BIN_FILE:
                    img.to_bin(self._replace_ext(f, ".bin"),
                               compress=self.compress)
                elif self.ofmt == OutputFormat.C_ARRAY:
                    img.to_c_array(self._replace_ext(f, ".c", outputname),
                                   compress=self.compress,
                                   outputname=outputname)
                elif self.ofmt == OutputFormat.PNG_FILE:
                    img.to_png(self._replace_ext(f, ".png"))

        return output


def main():
    parser = argparse.ArgumentParser(description='LVGL PNG to bin image tool.')
    parser.add_argument('--ofmt',
                        help="output filename format, C or BIN",
                        default="BIN",
                        choices=["C", "BIN", "PNG"])
    parser.add_argument(
        '--cf',
        help=("bin image color format, use AUTO for automatically "
              "choose from I1/2/4/8"),
        default="I8",
        choices=[
            "L8", "I1", "I2", "I4", "I8", "A1", "A2", "A4", "A8", "AL88", "ARGB8888",
            "XRGB8888", "RGB565", "RGB565_SWAPPED", "RGB565A8", "ARGB8565", "RGB888", "AUTO",
            "RAW", "RAW_ALPHA", "ARGB8888_PREMULTIPLIED"
        ])

    parser.add_argument('--rgb565dither', action='store_true',
                        help="use dithering to correct banding in gradients", default=False)

    parser.add_argument('--premultiply', action='store_true',
                        help="pre-multiply color with alpha", default=False)

    parser.add_argument('--compress',
                        help=("Binary data compress method, default to NONE"),
                        default="NONE",
                        choices=["NONE", "RLE", "LZ4"])

    parser.add_argument('--align',
                        help="stride alignment in bytes for bin image",
                        default=1,
                        type=int,
                        metavar='byte',
                        nargs='?')
    parser.add_argument('--background',
                        help="Background color for formats without alpha",
                        default=0x00_00_00,
                        type=lambda x: int(x, 0),
                        metavar='color',
                        nargs='?')
    parser.add_argument('--nemagfx', action='store_true',
                    help="export color palette for I8 images in a format compatible with NEMA accelerator", default=False)
    parser.add_argument('-o',
                        '--output',
                        default="./output",
                        help="Select the output folder, default to ./output")
    parser.add_argument('--name',
                        default=None,
                        help="Specify name for output file. Only applies when input is a file, not a directory. (Also used for variable name inside .c file when format is 'C')")
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument(
        'input', help="the filename or folder to be recursively converted")

    args = parser.parse_args()

    if path.isfile(args.input):
        files = [args.input]
    elif path.isdir(args.input):
        files = list(Path(args.input).rglob("*.[pP][nN][gG]"))

        if args.name is not None:
            raise BaseException(f"invalid input: cannot specify --name when input is a directory")
    else:
        raise BaseException(f"invalid input: {args.input}")

    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    logging.info(f"options: {args.__dict__}, files:{[str(f) for f in files]}")

    if args.cf == "AUTO":
        cf = None
    else:
        cf = ColorFormat[args.cf]

    ofmt = OutputFormat(args.ofmt) if cf not in (
        ColorFormat.RAW, ColorFormat.RAW_ALPHA) else OutputFormat.C_ARRAY
    compress = CompressMethod[args.compress]

    converter = PNGConverter(files,
                             cf,
                             ofmt,
                             args.output,
                             background=args.background,
                             align=args.align,
                             premultiply=args.premultiply,
                             compress=compress,
                             keep_folder=False,
                             rgb565_dither=args.rgb565dither,
                             nema_gfx=args.nemagfx)
    output = converter.convert(args.name)
    for f, img in output:
        logging.info(f"len: {img.data_len} for {path.basename(f)} ")

    print(f"done {len(files)} files")


def test():
    logging.basicConfig(level=logging.INFO)
    f = "pngs/cogwheel.RGB565A8.png"
    img = LVGLImage().from_png(f,
                               cf=ColorFormat.ARGB8565,
                               background=0xFF_FF_00,
                               rgb565_dither=True)
    img.adjust_stride(align=16)
    img.premultiply()
    img.to_bin("output/cogwheel.ARGB8565.bin")
    img.to_c_array("output/cogwheel-abc.c")  # file name is used as c var name
    img.to_png("output/cogwheel.ARGB8565.png.png")  # convert back to png


def test_raw():
    logging.basicConfig(level=logging.INFO)
    f = "pngs/cogwheel.RGB565A8.png"
    img = RAWImage().from_file(f,
                               cf=ColorFormat.RAW_ALPHA)
    img.to_c_array("output/cogwheel-raw.c")


if __name__ == "__main__":
    # test()
    # test_raw()
    main()


# emit/c_file.py

class CFile:
    def __init__(self, header_filename):
        self.lines = []
        self.header_filename = header_filename

        self._emit_preamble()

    def _emit_preamble(self):
        self.add(f'#include "{self.header_filename}"')
        self.add('#include "ui_defs.h"')
        self.add("")

    def add(self, line=""):
        self.lines.append(line)

    def add_block(self, title, lines):
        self.add("")
        self.add("// ------------------------------")
        self.add(f"// {title}")
        self.add("// ------------------------------")
        for line in lines:
            self.add(line)

    def render(self):
        return "\n".join(self.lines) + "\n"



import sys
import os
import re
import xml.etree.ElementTree as ET

# ---------------------------
# Configuration / constants
# ---------------------------
UI_MAX_STRING_LENGTH = 64
UI_MAX_CHILDREN = 64  # if you want, adjust this
INDENT = "    "

# Mapping simple tags -> child types
# "Text" -> label
# tags starting with "icon_" -> icon
# If you want to add more mappings, change here.

# ---------------------------
# Helpers
# ---------------------------


def normalize_id(name):
    return name.lower().replace("-", "_").replace(" ", "_")

def to_snake_case(s):
    # normalize whitespace and punctuation, then to snake_case
    s = s.strip()
    # replace non-alnum with underscore
    s = re.sub(r'[^0-9a-zA-Z]+', '_', s)
    # collapse multiple underscores
    s = re.sub(r'__+', '_', s)
    return s.lower().strip('_')

def base_name_for_header(frame_name_snake):
    # If frame name ends with "_screen", remove it for base
    if frame_name_snake.endswith("_screen"):
        return frame_name_snake[:-7]
    return frame_name_snake

def sanitize_c_string(s, maxlen=UI_MAX_STRING_LENGTH):
    if s is None:
        return ""
    s = s.replace('"', '\\"')
    if len(s) >= maxlen:
        return s[:maxlen-1]
    return s

def int_attr(node, key):
    v = node.attrib.get(key)
    if v is None or v == "":
        return 0
    try:
        return int(float(v))
    except:
        return 0

def write_file(path, text):
    with open(path, 'w', encoding='utf-8') as f:
        f.write(text)



# templates/label_templates.py


#LABEL_JOB_STRUCT = """
#typedef struct {
#    uint8_t child_index;
#    char text[UI_MAX_STRING_LENGTH];
#} {job_struct};
#"""

# REMOVED: LABEL_JOB_CALLBACK — no longer needed, logic moves inline to setter

LABEL_CALLBACK=""

LABEL_SETTER = """
void ${fn_name}(const char *text)
{
    ui_child_t *c = &${screen_var}.children[${child_index}];
    if (c->lv_obj) {
        lv_label_set_text(c->lv_obj, text);
    }
}
"""

LABEL_INIT = """
    case UI_CHILD_LABEL:
        c->lv_obj = lv_label_create(${screen_var}.lv_screen);
        lv_obj_set_pos(c->lv_obj, c->x, c->y);
        lv_obj_set_width(c->lv_obj, c->w);
        lv_label_set_long_mode(c->lv_obj, LV_LABEL_LONG_CLIP);
        break;
"""


import sys
import os
import shutil
import subprocess
import argparse
import xml.etree.ElementTree as ET
from pathlib import Path

from figma2lvgl.core.figma_parser import parse_screen
from figma2lvgl.core.generator import generate_screen
from figma2lvgl.core.utils.utils import write_file
from figma2lvgl.core.cmake_generator import generate_cmake
from figma2lvgl.core.child_registry import CHILDREN





#The image converion code downloaded and used from offical lvgl repo, has header inclusion
#which falss back to "lvgl/lvgl.h" which causes compile error for esp32. it relaces that
# block after generation
#
def fix_lvgl_includes(priv_src_dir):
    old_block = '''#if defined(LV_LVGL_H_INCLUDE_SIMPLE)
#include "lvgl.h"
#elif defined(LV_LVGL_H_INCLUDE_SYSTEM)
#include <lvgl.h>
#elif defined(LV_BUILD_TEST)
#include "../lvgl.h"
#else
#include "lvgl/lvgl.h"
#endif'''

    new_block = '''#if defined(LV_LVGL_H_INCLUDE_SIMPLE)
#include "lvgl.h"
#elif defined(LV_LVGL_H_INCLUDE_SYSTEM)
#include <lvgl.h>
#elif defined(LV_BUILD_TEST)
#include "../lvgl.h"
#elif defined(ESP_PLATFORM)
#include "lvgl.h"
#else
#include "lvgl/lvgl.h"
#endif'''

    for c_file in Path(priv_src_dir).glob("*.c"):
        with open(c_file, "r", encoding="utf-8") as f:
            content = f.read()

        if old_block in content:
            content = content.replace(old_block, new_block)

            with open(c_file, "w", encoding="utf-8") as f:
                f.write(content)

            print(f"  Patched LVGL include block in {c_file.name}")


# -------------------------------------------------
# Clean directory completely and recreate it
# -------------------------------------------------
# -------------------------------------------------
# Clean directory completely and recreate it
# -------------------------------------------------
def reset_directory(path):
    if os.path.exists(path):
        shutil.rmtree(path)
    os.makedirs(path, exist_ok=True)


# -------------------------------------------------
# Check if ui_src already has content and ask user
# -------------------------------------------------
def confirm_overwrite(ui_src: Path) -> bool:
    """
    Returns True if it's safe to proceed (either folder is fresh,
    or the user explicitly said yes).
    """
    if not ui_src.exists():
        return True  # nothing there yet, no need to ask

    # Check if any of the 4 subfolders exist AND have files in them
    subfolders = ["src", "include", "priv_src", "priv_include"]
    has_content = any(
        (ui_src / sub).is_dir() and any((ui_src / sub).iterdir())
        for sub in subfolders
    )

    if not has_content:
        return True  # folders exist but are empty, safe to proceed

    print("\n==========================================")
    print(" WARNING: Output folder already has content")
    print("==========================================")
    print(f"  {ui_src}")
    print("\n  The following subfolders will be wiped and regenerated:")
    for sub in subfolders:
        folder = ui_src / sub
        if folder.is_dir():
            count = len(list(folder.iterdir()))
            print(f"    {sub}/  ({count} file(s))")
    print()

    while True:
        answer = input("  Overwrite? [y/n]: ").strip().lower()
        if answer in ("y", "yes"):
            return True
        if answer in ("n", "no"):
            print("\nAborted. No files were changed.\n")
            return False
        print("  Please enter y or n.")

# -------------------------------------------------
# Validate required assets exist
# -------------------------------------------------
def validate_assets(screens, images_dir):
    required_assets = set()
    for screen in screens:
        required_assets.update(screen.get_required_assets(CHILDREN))

    missing = []
    for asset_id in required_assets:
        expected_file = os.path.join(images_dir, asset_id + ".png")
        if not os.path.exists(expected_file):
            missing.append(expected_file)

    if missing:
        print("\n==========================================")
        print(" ASSET VALIDATION FAILED")
        print("==========================================")
        print("The following image files are missing:\n")
        for m in missing:
            print("  -", m)
        print("\nPlease add the missing files and run again.")
        print("==========================================\n")
        return False

    return True


# -------------------------------------------------
# Run image converter script with dest folders
# -------------------------------------------------
def run_image_converter(images_dir, priv_src_dir, priv_include_dir, lvgl_tool):
    script = Path(__file__).parent / "tools" / "image_converter.py"
    print(f"DEBUG: Running script at -> {script}")

    if not script.exists():
        print(f"ERROR: image_converter.py not found at {script}")
        return False

    print("\nRunning image conversion...")

    result = subprocess.run(
        [
            sys.executable,
            str(script),
            str(images_dir),
            str(priv_src_dir),
            str(priv_include_dir),
            str(lvgl_tool)          # ← new 4th argument
        ],
        capture_output=True,
        text=True
    )

    if result.returncode != 0:
        print("\n==========================================")
        print(" IMAGE CONVERSION FAILED")
        print("==========================================")
        print(result.stdout)
        print(result.stderr)
        print("==========================================\n")
        return False

    print("Image conversion completed successfully.")
    return True
# -------------------------------------------------
# Argument parsing
# -------------------------------------------------
def parse_args():
    parser = argparse.ArgumentParser(
        description="LVGL UI Generator — converts Figma XML + images to C/H source files.",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=(
            "Examples:\n"
            "  Windows:  python main.py -x E:/project/diagram.xml\n"
            "  Linux:    python main.py -x /home/user/project/diagram.xml\n"
            "  Full:     python main.py -x diagram.xml -i ./assets/images -d ./output\n"
        )
    )

    parser.add_argument(
        "-x", "--xml",
        required=True,
        metavar="PATH",
        help="(Required) Path to the diagram XML file."
    )

    parser.add_argument(
        "-i", "--images",
        metavar="PATH",
        default=None,
        help=(
            "Path to the folder containing PNG images.\n"
            "Default: same directory as the XML file."
        )
    )

    parser.add_argument(
        "-d", "--dest",
        metavar="PATH",
        default=None,
        help=(
            "Destination folder where ui_src/ will be created.\n"
            "Default: same directory as the XML file.\n"
            "         This is recommended — output stays next to your project."
        )
    )

    return parser.parse_args()

# -------------------------------------------------
# Copy static files into priv_include
# -------------------------------------------------
def copy_static_files(priv_src: Path, priv_inc: Path) -> bool:
    static_src = Path(__file__).parent / "static_src"

    if not static_src.is_dir():
        print(f"WARNING: static_src/ not found at {static_src}, skipping.")
        return True

    copied = 0

    # Copy headers → priv_include
    for header in static_src.glob("*.h"):
        shutil.copy2(str(header), str(priv_inc / header.name))
        print(f"  Copied {header.name} -> priv_include/")
        copied += 1

    # Copy sources → priv_src
    for source in static_src.glob("*.c"):
        shutil.copy2(str(source), str(priv_src / source.name))
        print(f"  Copied {source.name} -> priv_src/")
        copied += 1

    if copied == 0:
        print("WARNING: static_src/ exists but contains no .h/.c files.")

    return True

#--Dependencies
def find_or_download_lvgl_tool() -> Path:
    # Standard user cache directory, cross-platform
    import platformdirs
    cache_dir = Path(platformdirs.user_cache_dir("figma2lvgl"))
    cached = cache_dir / "LVGLImage.py"

    if cached.is_file():
        return cached

    print("\n  figma2lvgl requires LVGLImage.py from the LVGL project (MIT licensed).")
    print("  Source: https://github.com/lvgl/lvgl/blob/master/scripts/LVGLImage.py")
    print()

    answer = input("  Download and cache it now? [y/n]: ").strip().lower()
    if answer not in ("y", "yes"):
        print("  Aborted. You can manually place LVGLImage.py next to image_converter.py.")
        return None

    import urllib.request
    url = "https://raw.githubusercontent.com/lvgl/lvgl/master/scripts/LVGLImage.py"
    cache_dir.mkdir(parents=True, exist_ok=True)

    print("  Downloading...")
    urllib.request.urlretrieve(url, cached)
    print(f"  Saved to {cached}")
    return cached

#############
# -------------------------------------------------
# Main
# -------------------------------------------------
def main():
    args = parse_args()

    # --- Resolve XML path ---
    xml_path = Path(args.xml).resolve()
    if not xml_path.is_file():
        print(f"ERROR: XML file not found: {xml_path}")
        sys.exit(1)

    # --- Resolve images dir (default: cwd) ---
    images_dir = Path(args.images).resolve() if args.images else xml_path.parent
    if not images_dir.is_dir():
        print(f"ERROR: Images directory not found: {images_dir}")
        sys.exit(1)
    
    

    # --- Resolve destination dir (default: folder containing the XML) ---
    dest_dir = Path(args.dest).resolve() if args.dest else xml_path.parent

    # --- Build output folder tree ---
    #
    #  <dest>/ui_src/
    #    src/           ← .c files generated from XML
    #    include/       ← .h files generated from XML
    #    priv_src/      ← .c files from image converter
    #    priv_include/  ← .h files from image converter
    #
    ui_src      = dest_dir  / "ui_src"
    src_dir     = ui_src    / "src"
    include_dir = ui_src    / "include"
    priv_src    = ui_src    / "priv_src"
    priv_inc    = ui_src    / "priv_include"

    print("\n==========================================")
    print(" LVGL UI Generator")
    print("==========================================")
    print(f"  XML file   : {xml_path}")
    print(f"  Images dir : {images_dir}")
    print(f"  Output dir : {ui_src}")
    print("==========================================\n")

    # --- Parse XML ---
    try:
        tree = ET.parse(xml_path)
        root = tree.getroot()
    except Exception as e:
        print(f"ERROR: Failed to parse XML: {e}")
        sys.exit(1)

    page_children = root.find("children")
    if page_children is None:
        print("ERROR: No <children> element found in XML.")
        sys.exit(1)

    frames = page_children.findall("Frame")
    if not frames:
        print("ERROR: No <Frame> nodes found in XML.")
        sys.exit(1)

    screens = [parse_screen(frame) for frame in frames]

    # --- Validate assets ---
    if not validate_assets(screens, images_dir):
        sys.exit(1)

    lvgl_tool = find_or_download_lvgl_tool()
    if lvgl_tool is None:
        sys.exit(1)

    # --- Reset output folders ---
      # --- Confirm overwrite if ui_src already exists ---
    if not confirm_overwrite(ui_src):
        sys.exit(0)

    # --- Reset output folders ---
    print("Cleaning output folders...")
    for folder in [src_dir, include_dir, priv_src, priv_inc]:
        reset_directory(folder)

    # --- Run image converter ---
    if not run_image_converter(images_dir, priv_src, priv_inc, lvgl_tool):
            sys.exit(1)
    
    # --- Fix LVGL includes in converted files ---
    fix_lvgl_includes(priv_src)

     # --- Generate UI source files ---

    #---Copy static files, so both headers and src
    print("\nCopying static files...")
    if not copy_static_files(priv_src, priv_inc):
        sys.exit(1)

    # --- Generate UI source files ---
    print("\nGenerating UI files...\n")
    generated = []

    for screen in screens:
        c_fname, h_fname, h_text, c_text = generate_screen(screen)

        c_path = src_dir     / c_fname
        h_path = include_dir / h_fname

        write_file(str(h_path), h_text)
        write_file(str(c_path), c_text)
        generated.append((c_path, h_path))

    print("Generated files:")
    for c, h in generated:
        print(f"  - {c}")
        print(f"  - {h}")

    # --- Generate CMake ---
    #cmake_text = generate_cmake()
    #cmake_path = ui_src / "CMakeLists.txt"
    #write_file(str(cmake_path), cmake_text)

    print("\n==========================================")
    print(" PIPELINE COMPLETED SUCCESSFULLY")
    print("==========================================")
    print(f"\n  ui_src/")
    print(f"    src/          ({len(list(src_dir.glob('*.c')))} .c files)")
    print(f"    include/      ({len(list(include_dir.glob('*.h')))} .h files)")
    print(f"    priv_src/     ({len(list(priv_src.glob('*.c')))} .c files)")
    print(f"    priv_include/ ({len(list(priv_inc.glob('*.h')))} .h files)")
    print()


if __name__ == "__main__":
    main()


# core/generator.py

from figma2lvgl.core.child_registry import CHILDREN
from figma2lvgl.core.emit.c_file import CFile
from figma2lvgl.core.emit.h_file import HFile
from figma2lvgl.core.emit.layouts import C_FILE_LAYOUT, H_FILE_LAYOUT
from figma2lvgl.core.utils.utils import to_snake_case
from figma2lvgl.core.utils.template_loader import load_template
from string import Template


# ── Style block renderer ──────────────────────────────────────────────────────

_ALIGN_MAP = {
    "LEFT":   "LV_TEXT_ALIGN_LEFT",
    "CENTER": "LV_TEXT_ALIGN_CENTER",
    "RIGHT":  "LV_TEXT_ALIGN_RIGHT",
}

def _format_sub_struct(name, fields, indent) -> str:
    """
    Render one sub-struct with multiline field formatting.

    Result:
        .box = {
            .has_bg = true,
            .bg = lv_color_hex(0xFFFFFF),
        }
    """
    field_indent = indent + "    "
    fields_str   = (",\n" + field_indent).join(fields)
    return (
        f".{name} = {{\n"
        f"{field_indent}{fields_str}\n"
        f"{indent}}}"
    )


def _render_style_block(style, indent="            ") -> str:
    """
    Convert a ParsedStyle into a C struct initializer block.
    Emits .style = {0} when no properties were found.

    indent — base indentation of the .style field in the generated file.
             Passed in so this function never hardcodes whitespace.
    """
    if style.is_empty():
        return ".style = { .box = { 0 }, .text = { 0 }, .effects = { 0 } },"

    # ── box ──────────────────────────────────────────────────────────────────
    box = style.box
    box_fields = []
    if box.bg_color is not None:
        box_fields.append(f".has_bg = true")
        box_fields.append(f".bg = 0x{box.bg_color:06X}")
    if box.bg_opa is not None:
        box_fields.append(f".has_bg_opa = true")
        box_fields.append(f".bg_opa = {box.bg_opa}")
    if box.border_color is not None:
        box_fields.append(f".has_border_color = true")
        box_fields.append(f".border_color = 0x{box.border_color:06X}")
    if box.border_width is not None:
        box_fields.append(f".has_border_width = true")
        box_fields.append(f".border_width = {box.border_width}")
    if box.radius is not None:
        box_fields.append(f".has_radius = true")
        box_fields.append(f".radius = {box.radius}")

    # ── text ─────────────────────────────────────────────────────────────────
    text = style.text
    text_fields = []
    if text.color is not None:
        text_fields.append(f".has_color = true")
        text_fields.append(f".color = 0x{text.color:06X}")
    if text.size is not None:
        text_fields.append(f".has_size = true")
        text_fields.append(f".size = {text.size}")
    if text.align is not None:
        text_fields.append(f".has_align = true")
        text_fields.append(f".align = {_ALIGN_MAP[text.align]}")

    # ── effects ───────────────────────────────────────────────────────────────
    effects = style.effects
    effects_fields = []
    if effects.opacity is not None:
        effects_fields.append(f".has_opacity = true")
        effects_fields.append(f".opacity = {effects.opacity}")

    # ── assemble ──────────────────────────────────────────────────────────────
    inner = []
    sub_indent = indent + "    "

    if box_fields:
        inner.append(_format_sub_struct("box", box_fields, sub_indent))
    if text_fields:
        inner.append(_format_sub_struct("text", text_fields, sub_indent))
    if effects_fields:
        inner.append(_format_sub_struct("effects", effects_fields, sub_indent))

    if not inner:
        return ".style = { 0 },"

    sections = (",\n" + sub_indent).join(inner)
    return (
        f".style = {{\n"
        f"{sub_indent}{sections}\n"
        f"{indent}}},"
    )


# ── Screen generator ──────────────────────────────────────────────────────────

def generate_screen(screen):


    print("DEBUG: generate_screen is running")
    screen_snake    = screen.snake
    base            = screen_snake
    header_filename = f"ui_{base}.h"
    guard           = f"UI_{base.upper()}_H"

    init_fn  = f"ui_{base}_init"
    load_fn  = f"ui_{base}_load"
    load_cb  = f"ui_{base}_load_job"

    # --------------------------
    # Build screen struct
    # --------------------------

    child_entries = []

    for child in screen.children:

        data_block = ""

        if child.type == "UI_CHILD_LABEL":
            data_block = """
                .data.label = {
                    .text = ""
                }
    """
        elif child.type == "UI_CHILD_IMAGE":
            data_block = """
                .data.image = {
                    .src = NULL
                }
    """
        elif child.type == "UI_CHILD_BAR":
            data_block = """
                .data.bar = {
                    .value = 0
                }
    """

        style_block = _render_style_block(child.style)

        entry = f"""
            {{
                .type = {child.type},
                .id = "{child.id}",
                .lv_obj = NULL,
                .x = {child.x},
                .y = {child.y},
                .w = {child.w},
                .h = {child.h},
                {style_block}
        {data_block}
            }},
        """
        child_entries.append(entry)

    screen_struct = f"""
    ui_screen_t {screen_snake} = {{
        .name = "{screen.name}",
        .child_count = {len(screen.children)},
        .children = {{
            {"".join(child_entries)}
        }},
        .lv_screen = NULL
    }};
    """

    # --------------------------
    # Callbacks / setters
    # --------------------------

    job_callbacks    = []
    setters          = []
    setter_prototypes = []

    unique_types = set()

    # --------------------------
    # CHILD LOOP → setters only
    # --------------------------

    for index, child in enumerate(screen.children):

        spec = CHILDREN.get(child.type)
        if not spec:
            continue

        unique_types.add(child.type)

        if child.type == "UI_CHILD_LABEL":
            cb_name = ""
            fn_name = f"ui_{screen_snake}_set_{child.id}"

        elif child.type == "UI_CHILD_IMAGE":
            cb_name = ""
            fn_name = f"ui_{screen_snake}_display_{child.id}"

        elif child.type == "UI_CHILD_BAR":
            cb_name = f"ui_{screen_snake}_bar_job_cb"
            fn_name = f"ui_{screen_snake}_set_{child.id}"

        else:
            continue

        setter_tpl = load_template(spec.setter_template)

        setters.append(
            Template(setter_tpl).safe_substitute(
                fn_name=fn_name,
                child_index=index,
                screen_var=screen_snake,
                child_id=child.id,
                cb_name=cb_name
            )
        )

        setter_prototypes.append(
            f"void {fn_name}({spec.setter_args});"
        )

    # --------------------------
    # TYPE LOOP → callbacks only
    # --------------------------

    for type_name in unique_types:

        spec = CHILDREN.get(type_name)
        if not spec:
            continue

        if type_name == "UI_CHILD_LABEL":
            cb_name = ""
        elif type_name == "UI_CHILD_IMAGE":
            cb_name = ""
        elif type_name == "UI_CHILD_BAR":
            cb_name = f"ui_{screen_snake}_bar_job_cb"
        else:
            continue

        callback_tpl = load_template(spec.callback_template)

        job_callbacks.append(
            Template(callback_tpl).safe_substitute(
                cb_name=cb_name,
                screen_var=screen_snake
            )
        )

    # --------------------------
    # Init cases (ONE PER TYPE)
    # --------------------------

    init_cases = []

    for type_name in unique_types:

        spec = CHILDREN.get(type_name)
        if not spec:
            continue

        init_tpl = load_template(spec.init_template)

        init_cases.append(
            Template(init_tpl).safe_substitute(
                screen_var=screen_snake
            )
        )

    # --------------------------
    # Assemble C file
    # --------------------------

    c_text = Template(C_FILE_LAYOUT).safe_substitute(
        header_filename=header_filename,
        screen_struct=screen_struct,
        job_callbacks="\n".join(job_callbacks),
        setters="\n".join(setters),
        sc_fn_cb_name=load_cb,
        sc_fn_name=load_fn,
        init_fn=init_fn,
        screen_var=screen_snake,
        init_body="\n".join(init_cases)
    )

    # --------------------------
    # Assemble H file
    # --------------------------

    h_text = Template(H_FILE_LAYOUT).safe_substitute(
        guard=guard,
        init_fn=init_fn,
        sc_fn_name=load_fn,
        setter_prototypes="\n".join(setter_prototypes)
    )

    return (
        f"ui_{base}.c",
        header_filename,
        h_text,
        c_text
    )



# core/utils/template_loader.py

from figma2lvgl.core.templates import (
    label_templates,
    image_templates,
    bar_templates,
)

TEMPLATE_MAP = {
    # LABEL
    "label_callback": label_templates.LABEL_CALLBACK,
    "label_setter": label_templates.LABEL_SETTER,
    "label_init": label_templates.LABEL_INIT,

    # IMAGE
    "image_callback": image_templates.IMAGE_CALLBACK,
    "image_setter": image_templates.IMAGE_SETTER,
    "image_init": image_templates.IMAGE_INIT,

    # BAR
    "bar_callback": bar_templates.BAR_CALLBACK,
    "bar_setter": bar_templates.BAR_SETTER,
    "bar_init": bar_templates.BAR_INIT,
}


def load_template(name):
    if not name:
        return ""
    return TEMPLATE_MAP[name]


# code_buffer.py

class CodeBuffer:
    def __init__(self):
        self._lines = []

    def add(self, line=""):
        self._lines.append(line)

    def extend(self, lines):
        self._lines.extend(lines)

    def render(self):
        return "\n".join(self._lines) + "\n"



# model/screen.py

from model.child import Child

class Screen:
    """
    Pure semantic representation of a screen.
    """

    def __init__(self, name):
        self.name = name
        self.children = []     # list[Child]

    def add_child(self, child: Child):
        self.children.append(child)

    def __repr__(self):
        return f"Screen(name={self.name}, children={len(self.children)})"



# model/child.py

class Child:
    """
    Pure semantic representation of a UI child.
    No XML, no LVGL, no C.
    """

    def __init__(self, *, type, id, x, y, w, h):
        self.type = type        # "UI_CHILD_LABEL", "UI_CHILD_ICON", ...
        self.id = id            # normalized id (string)
        self.x = x
        self.y = y
        self.w = w
        self.h = h

    def __repr__(self):
        return (
            f"Child(type={self.type}, id={self.id}, "
            f"x={self.x}, y={self.y}, w={self.w}, h={self.h})"
        )





def map_tag_to_child_type(node):

    name = node.attrib.get("name", "").lower()

    if node.tag == "Text":
        return "UI_CHILD_LABEL"

    if "bar" in name:
        return "UI_CHILD_BAR"

    if "icon" in name or "image" in name:
        return "UI_CHILD_IMAGE"

    return "UI_CHILD_LABEL"



def int_attr(node, key):
    v = node.attrib.get(key)
    if v is None or v == "":
        return 0
    try:
        return int(float(v))
    except:
        return 0



# emit/h_file.py

class HFile:
    def __init__(self, guard):
        self.guard = guard
        self.lines = []

        self._emit_preamble()

    def _emit_preamble(self):
        self.lines.append(f"#ifndef {self.guard}")
        self.lines.append(f"#define {self.guard}")
        self.lines.append("")
        self.lines.append("#ifdef __cplusplus")
        self.lines.append('extern "C" {')
        self.lines.append("#endif")
        self.lines.append("")

    def add(self, line=""):
        self.lines.append(line)

    def add_prototype(self, proto):
        self.lines.append(proto)

    def close(self):
        self.lines.append("")
        self.lines.append("#ifdef __cplusplus")
        self.lines.append("}")
        self.lines.append("#endif")
        self.lines.append("")
        self.lines.append("#endif")

    def render(self):
        return "\n".join(self.lines) + "\n"



from figma2lvgl.core.config_name import runtime_folder_name

def generate_cmake():
    runtime_dir = runtime_folder_name()

    return f"""
file(GLOB GENERATED_SOURCES
    "${{CMAKE_CURRENT_LIST_DIR}}/src_generated/*.c"
)

file(GLOB RUNTIME_SOURCES
    "${{CMAKE_CURRENT_LIST_DIR}}/runtime/*.c"
)

file(GLOB GENERATED_ASSET_SOURCES
    "${{CMAKE_CURRENT_LIST_DIR}}/assets/src_generated/*.c"
)



idf_component_register(
    SRCS
        ${{GENERATED_SOURCES}}
        ${{RUNTIME_SOURCES}}
        ${{GENERATED_ASSET_SOURCES}}
    INCLUDE_DIRS
        "src_generated"
    PRIV_INCLUDE_DIRS
        "runtime"
        "assets/src_generated"
    PRIV_REQUIRES
        lvgl
        esp_lvgl_port
)

"""



# figma_parser.py

from figma2lvgl.core.utils.utils import normalize_id, to_snake_case
from figma2lvgl.core.utils.figma_helpers import map_tag_to_child_type, int_attr


# ── Color helpers ─────────────────────────────────────────────────────────────

def _hex_to_int(hex_color: str):
    """
    "#d9d9d9" → 0xd9d9d9
    Returns None if invalid or missing.
    """
    if not hex_color:
        return None
    hex_color = hex_color.strip().lstrip("#")
    try:
        return int(hex_color, 16)
    except ValueError:
        return None


def _first_visible_fill(node):
    """
    Returns the first <fill> under <fills> that is visible.
    Skips fills with visible="false".
    """
    fills = node.find("fills")
    if fills is None:
        return None
    for fill in fills.findall("fill"):
        if fill.attrib.get("visible", "true").lower() == "false":
            continue
        return fill
    return None


def _first_stroke(node):
    """Returns the first <stroke> under <strokes>."""
    strokes = node.find("strokes")
    if strokes is None:
        return None
    return strokes.find("stroke")


# ── Style data classes ────────────────────────────────────────────────────────

class ParsedStyleBox:
    def __init__(self):
        self.bg_color       = None   # int hex e.g. 0xFFFFFF, or None
        self.bg_opa         = None   # 0-255 or None
        self.border_color   = None   # int hex or None
        self.border_width   = None   # int or None
        self.radius         = None   # int or None


class ParsedStyleText:
    def __init__(self):
        self.color          = None   # int hex or None
        self.size           = None   # int or None
        self.align          = None   # "LEFT" / "CENTER" / "RIGHT" or None


class ParsedStyleEffects:
    def __init__(self):
        self.opacity        = None   # 0-255 or None


class ParsedStyle:
    def __init__(self):
        self.box     = ParsedStyleBox()
        self.text    = ParsedStyleText()
        self.effects = ParsedStyleEffects()

    def is_empty(self):
        """Returns True if no style properties were found — emit .style = {0}"""
        return all([
            self.box.bg_color       is None,
            self.box.bg_opa         is None,
            self.box.border_color   is None,
            self.box.border_width   is None,
            self.box.radius         is None,
            self.text.color         is None,
            self.text.size          is None,
            self.text.align         is None,
            self.effects.opacity    is None,
        ])


# ── Style extraction ──────────────────────────────────────────────────────────

def parse_style(node, child_type) -> ParsedStyle:
    """
    Type-aware style extraction from a Figma XML node.

    node        — the XML element
    child_type  — mapped UI_CHILD_* string e.g. "UI_CHILD_LABEL"
    """
    style     = ParsedStyle()
    node_type = node.attrib.get("type", "")

    # Infer text context from either Figma node type or mapped child type
    # so labels are handled correctly even if node type attribute is absent
    is_text = (node_type == "TEXT") or (child_type == "UI_CHILD_LABEL")

    # ── Fill color ────────────────────────────────────────────────────────────
    fill = _first_visible_fill(node)
    if fill is not None:
        color_int = _hex_to_int(fill.attrib.get("color"))
        if color_int is not None:
            if is_text:
                # fills on TEXT nodes = text color, not background
                style.text.color = color_int
            else:
                style.box.bg_color = color_int

        # fill-level opacity e.g. <fill opacity="0.5" ... />
        fill_opa = fill.attrib.get("opacity")
        if fill_opa is not None and not is_text:
            try:
                style.box.bg_opa = round(float(fill_opa) * 255)
            except ValueError:
                pass

    # ── Stroke / border ───────────────────────────────────────────────────────
    stroke = _first_stroke(node)
    if stroke is not None:
        color_int = _hex_to_int(stroke.attrib.get("color"))
        if color_int is not None:
            style.box.border_color = color_int

    stroke_weight = node.attrib.get("strokeWeight")
    if stroke_weight is not None:
        try:
            style.box.border_width = round(float(stroke_weight))
        except ValueError:
            pass

    # if color was found but no width, default to 1 so border is visible in LVGL
    if style.box.border_color is not None and style.box.border_width is None:
        style.box.border_width = 1

    # ── Corner radius ─────────────────────────────────────────────────────────
    radius = node.attrib.get("cornerRadius")
    if radius is not None:
        try:
            style.box.radius = round(float(radius))
        except ValueError:
            pass

    # ── Text-only properties ──────────────────────────────────────────────────
    if is_text:
        font_size = node.attrib.get("fontSize")
        if font_size is not None:
            try:
                style.text.size = round(float(font_size))
            except ValueError:
                pass

        align = node.attrib.get("textAlignHorizontal")
        if align in ("LEFT", "CENTER", "RIGHT"):
            style.text.align = align

    # ── Whole-widget opacity ──────────────────────────────────────────────────
    opacity = node.attrib.get("opacity")
    if opacity is not None:
        try:
            style.effects.opacity = round(float(opacity) * 255)
        except ValueError:
            pass

    return style


# ── Parsed data classes ───────────────────────────────────────────────────────

class ParsedChild:
    def __init__(self, type, id, x, y, w, h, style=None):
        self.type   = type
        self.id     = id
        self.x      = x
        self.y      = y
        self.w      = w
        self.h      = h
        self.style  = style or ParsedStyle()


class ParsedScreen:
    def __init__(self, name):
        self.name     = name
        self.snake    = to_snake_case(name)
        self.children = []

    def get_required_assets(self, child_registry):
        assets = []
        for child in self.children:
            spec = child_registry.get(child.type)
            if spec and getattr(spec, "requires_asset", False):
                assets.append(child.id)
        return assets


# ── Screen parser ─────────────────────────────────────────────────────────────

def parse_screen(frame_node):
    frame_name = frame_node.attrib.get("name", "unnamed")
    screen     = ParsedScreen(frame_name)

    children_parent = frame_node.find("children")
    if children_parent is None:
        return screen

    for child in list(children_parent):
        mapped = map_tag_to_child_type(child)
        if mapped is None:
            continue

        x = int_attr(child, "x")
        y = int_attr(child, "y")
        w = int_attr(child, "width")
        h = int_attr(child, "height")

        raw_id   = child.attrib.get("name", f"child_{len(screen.children)}")
        child_id = normalize_id(raw_id)

        existing_ids = {c.id for c in screen.children}
        if child_id in existing_ids:
            raise ValueError(
                f"Duplicate child id '{child_id}' in screen '{frame_name}'"
            )

        style = parse_style(child, mapped)

        screen.children.append(
            ParsedChild(
                type=mapped,
                id=child_id,
                x=x, y=y, w=w, h=h,
                style=style
            )
        )

    return screen


# core/generic_child.py

class ChildSpec:
    def __init__(
        self,
        *,
        type_name,
        callback_template,
        setter_template,
        init_template,
        setter_args,
        requires_asset=False,
    ):
        self.type_name = type_name
        self.callback_template = callback_template
        self.setter_template = setter_template
        self.init_template = init_template
        self.setter_args = setter_args
        self.requires_asset = requires_asset


# child_registry.py

from figma2lvgl.core.generic_child import ChildSpec


CHILDREN = {
    "UI_CHILD_LABEL": ChildSpec(
        type_name="UI_CHILD_LABEL",
        #job_template="label_job",
        callback_template="",
        setter_template="label_setter",
        init_template="label_init",
        setter_args="const char *text",
    ),

    "UI_CHILD_IMAGE": ChildSpec(
    type_name="UI_CHILD_IMAGE",
    callback_template="",
    setter_template="image_setter",
    init_template="image_init",
    setter_args="void",
    requires_asset=True,   # NEW, main needs to know the names of image files and check if those named files exist 
    ),

    "UI_CHILD_BAR": ChildSpec(
        type_name="UI_CHILD_BAR",
        #job_template="bar_job",
        callback_template="bar_callback",
        setter_template="bar_setter",
        init_template="bar_init",
        setter_args="int value, uint32_t duration_ms"
    ),
}



import subprocess
import sys
import shutil
from pathlib import Path

COLOR_FORMAT = "RGB565"


def main():
    if len(sys.argv) != 5:
        print("Usage: python image_converter.py <images_dir> <priv_src_dir> <priv_include_dir> <lvgl_tool_path>")
        print("  This script is normally called by main.py automatically.")
        sys.exit(1)

    assets_dir = Path(sys.argv[1]).resolve()
    src_dir    = Path(sys.argv[2]).resolve()
    inc_dir    = Path(sys.argv[3]).resolve()
    lvgl_tool  = Path(sys.argv[4]).resolve()

    if not assets_dir.is_dir():
        print(f"ERROR: Images directory not found: {assets_dir}")
        sys.exit(1)

    if not lvgl_tool.is_file():
        print(f"ERROR: LVGLImage.py not found at: {lvgl_tool}")
        sys.exit(1)

    src_dir.mkdir(parents=True, exist_ok=True)
    inc_dir.mkdir(parents=True, exist_ok=True)

    staging_dir = src_dir.parent / "_lvgl_staging"
    staging_dir.mkdir(parents=True, exist_ok=True)

    image_names = []

    png_files = list(assets_dir.glob("*.png"))
    if not png_files:
        print(f"WARNING: No PNG files found in {assets_dir}")

    for png in png_files:
        print(f"Converting {png.name} ...")
        cmd = [
            sys.executable,
            str(lvgl_tool),       # ← use passed path, no guessing
            "--ofmt", "C",
            "--cf", COLOR_FORMAT,
            "-o", str(staging_dir),
            str(png)
        ]
        subprocess.run(cmd, check=True)
        image_names.append(png.stem)

    for file in staging_dir.iterdir():
        if file.suffix == ".c":
            shutil.move(str(file), str(src_dir / file.name))
        elif file.suffix == ".h":
            shutil.move(str(file), str(inc_dir / file.name))

    staging_dir.rmdir()

    generate_assets_header(image_names, inc_dir)
    print(f"\nDone.\n  Sources -> {src_dir}\n  Headers -> {inc_dir}")


def generate_assets_header(names, inc_dir):
    header = inc_dir / "assets.h"
    with open(header, "w") as f:
        f.write("#pragma once\n\n")
        f.write('#include "lvgl.h"\n\n')
        for name in names:
            f.write(f"LV_IMG_DECLARE({name});\n")
    print(f"Generated assets.h with {len(names)} image(s)")


if __name__ == "__main__":
    main()
r'''

---

### Flow summary
```
main.py
  find_or_download_lvgl_tool()  ← asks user once, caches in platformdirs
        ↓ lvgl_tool path
  run_image_converter(..., lvgl_tool)
        ↓ passes as sys.argv[4]
  image_converter.py
        ↓ uses it directly, no searching
  LVGLImage.py
  '''